<?php

namespace Ems\Services;

use Ems\Repository\{ LancamentoLoteRepository, PeriodoRepository };
use Ems\Services\MarcacaoService;
use Ems\Services\VinculoService;
use Ems\Util\{ ArrayUtil, Utilities };
use Util\Environment;
use Util\NovoCodigo;
use Doctrine\ORM\EntityManager;

class LancamentoLoteService {
    /** @var LancamentoLoteRepository $lancamentoLoteRepository */
    protected $lancamentoLoteRepository;

    /** @var Environment $environment */
    protected $environment;

    /** @var MarcacaoService $environment */
    protected $marcacaoService;

    /** @var PeriodoRepository $environment */
    protected $periodoRepository;

    public function __construct(
        Environment $environment,
        LancamentoLoteRepository $lancamentoLoteRepository,
        MarcacaoService $marcacaoService,
        PeriodoRepository $periodoRepository,
        EstruturaService $estruturaService,
        VinculoService $vinculoService
    ) {

        $this->environment = $environment;
        $this->lancamentoLoteRepository = $lancamentoLoteRepository;
        $this->marcacaoService = $marcacaoService;
        $this->periodoRepository = $periodoRepository;
        $this->estruturaService = $estruturaService;
        $this->vinculoService = $vinculoService;
    }

    public function getLotesLancamento(
        string $cdOperador,
        int $nrPeriodo
    ): array {

        $nrOrg = $this->environment->getNrOrg();

        $periodo = $this->periodoRepository->getPeriodo($nrOrg, $nrPeriodo);

        $iniPeriodo = $periodo['DTINICIALAPURACAO'];
        $iniPeriodo = Utilities::formatDate($iniPeriodo); // Remove o "00:00:00" da data.

        $fimPeriodo = $periodo['DTFINALAPURACAO'];
        $fimPeriodo = Utilities::formatDate($fimPeriodo); // Remove o "00:00:00" da data.

        $params = [
            'P_NRORG'         => $nrOrg,
            'P_CDOPERADOR'    => $cdOperador,
            'P_INIPERIODO'    => $iniPeriodo,
            'P_FIMPERIODO'    => $fimPeriodo
        ];

        $lotes = $this->lancamentoLoteRepository->getLotesLancamento($params);

        return $lotes;
    }

    public function getLancamentosEvento(
        int $nrLote,
        int $nrEstrutura
    ): array {

        $nrOrg = (int) $this->environment->getNrOrg();

        $this->createAllVinculoEventoItems($nrLote, $nrEstrutura);

        $lote = $this->lancamentoLoteRepository->findOneLote($nrOrg, $nrLote);
        $dtCompetencia = $lote['DTCOMPETLOTEEVE'];
        $dtCompetencia = Utilities::formatDate($dtCompetencia); // Remove o "00:00:00" da data.

        $params = [
            'P_NRORG'         => $nrOrg,
            'P_NRORGPADRAO'   => 0,
            'P_NRLOTE'        => $nrLote
        ];

        $lancamentos = $this->lancamentoLoteRepository->getLancamentosEvento($params);

        return $lancamentos;
    }

    /**
     * Creates all missing vinculo-evento relations in FPA_ITEMLOTELANC
     */
    public function createAllVinculoEventoItems(
        int $nrLote,
        int $nrEstrutura
    ) {
        $nrOrg = $this->environment->getNrOrg();
        $cdOperador = $this->environment->getCdOperador();
        $vinculos = $this->vinculoService->getVinculosGestor($nrEstrutura);

        $nrVinculos = array_map(function($vinculo) {
            return $vinculo['NRVINCULOM'];
        }, $vinculos);

        foreach ($nrVinculos as $nrVinculo) {
            $nrEventos = $this->getMissingEventosInVinculoLote($nrLote, $nrVinculo);

            foreach ($nrEventos as $nrEvento) {
                $nrNovoItem = NovoCodigo::geraCodigo($nrOrg, 'FPA_ITEMLOTELANC', 1, 22, true);

                $params = [
                    'P_NRORG'      => $nrOrg,
                    'P_NRVINCULO'  => $nrVinculo,
                    'P_NRLOTE'     => $nrLote,
                    'P_NREVENTO'   => $nrEvento,
                    'P_NRITEM'     => $nrNovoItem,
                    'P_CDOPERADOR' => $cdOperador
                ];

                $this->lancamentoLoteRepository->insertItemLote($params);
            }
        }
    }

    /**
     * Gets the number of all events that doesn't have a item with the
     * specified "vinculo" in FPA_ITEMLOTELANC.
     */
    public function getMissingEventosInVinculoLote(
        int $nrLote,
        int $nrVinculo
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();
        $nrOrgPadrao = (int) $this->environment->getNrOrgPadrao();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRORGPADRAO' => $nrOrgPadrao,
            'P_NRLOTE'      => $nrLote,
            'P_NRVINCULO'   => $nrVinculo
        ];

        $eventos = $this->lancamentoLoteRepository->
            getMissingEventosInVinculoLote($params);

        $nrEventos = array_map(function($evento) {
            return (int) $evento['NREVENTOM'];
        }, $eventos);

        return $nrEventos;
    }

    public function getTiposMovimento(): array {

        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'         => $nrOrg,
            'P_NRORGPADRAO'   => 0
        ];

        $tiposMov = $this->lancamentoLoteRepository->getTiposMovimento($params);

        return $tiposMov;
    }

    public function insertLoteLancamento(
        string $dsLote,
        string $dtOcorrencia,
        string $dtCompetencia,
        string $cdOperador,
        int $nrTipoMov
    ): bool {
        $nrOrg = (int) $this->environment->getNrOrg();
        $nrLote = NovoCodigo::geraCodigo($nrOrg, 'FPA_LOTELANCEVENTO', 1, 22, true);

        $params = [
            'P_NRORG'      => $nrOrg,
            'P_NRLOTE'     => $nrLote,
            'P_CDOPERADOR' => $cdOperador,
            'P_DSLOTE'     => $dsLote,
            'P_DTOCOR'     => $dtOcorrencia,
            'P_DTCOMP'     => $dtCompetencia,
            'P_NRTIPOMOV'  => $nrTipoMov
        ];

        $this->lancamentoLoteRepository->insertLoteLancamento($params);

        return true;
    }

    public function getEventosLancamento(
        int $nrEstrutura
    ): array {

        $nrOrg = $this->environment->getNrOrg();
        $nrOrgPadrao = $this->environment->getNrOrgPadrao();

        $estruturasInferioresEAtual = [];
        $this->estruturaService->
            getEstruturasInferioresRec($nrEstrutura, $estruturasInferioresEAtual, true);

        $estruturas = ArrayUtil::toStringSQL($estruturasInferioresEAtual);

        $params = [
            'P_NRORG'         => $nrOrg,
            'P_NRORGPADRAO'    => $nrOrgPadrao,
            'P_NRESTRUTURAS' => $estruturas
        ];

        $eventos = $this->lancamentoLoteRepository->getEventosLancamento($params);

        return $eventos;
    }

    public function editLoteLancamento(
        int $nrLote,
        string $dsLote,
        string $dtOcorrencia,
        string $dtCompetencia,
        string $cdOperador,
        int $nrTipoMov
    ): bool {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'      => $nrOrg,
            'P_NRLOTE'     => $nrLote,
            'P_CDOPERADOR' => $cdOperador,
            'P_DSLOTE'     => $dsLote,
            'P_DTOCOR'     => $dtOcorrencia,
            'P_DTCOMP'     => $dtCompetencia,
            'P_NRTIPOMOV'  => $nrTipoMov
        ];

        $this->lancamentoLoteRepository->editLoteLancamento($params);

        return true;
    }

    public function confirmLoteLancamento(
        int $nrLote
    ): bool {

        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'  => $nrOrg,
            'P_NRLOTE' => $nrLote
        ];

        $this->lancamentoLoteRepository->confirmLoteLancamento($params);

        return true;
    }

    public function editLancamentos(
        array $lancamentos,
        string $cdOperador
    ): bool {

        $nrOrg = (int) $this->environment->getNrOrg();

        foreach ($lancamentos as $lancamento) {
            $nrLancamento = (int) $lancamento['NRITLOTELANC'];

            $valor = isset($lancamento['VREVENTOLOTE']) ?
                (float) $lancamento['VREVENTOLOTE'] :
                null;

            $referencia = isset($lancamento['VRREFFOLHALOTE']) ?
                (float) $lancamento['VRREFFOLHALOTE'] :
                null;

            $params = [
                'P_NRORG'        => $nrOrg,
                'P_NRLANCAMENTO' => $nrLancamento,
                'P_CDOPERADOR'   => $cdOperador,
                'P_VALOR'        => $valor,
                'P_REFERENCIA'   => $referencia
            ];

            $this->lancamentoLoteRepository->editLancamento($params);
        }

        return true;
    }

    public function deleteLancamentos(
        int $nrLote
    ): bool {

        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRLOTE' => $nrLote,
            'P_NRORG'  => $nrOrg
        ];

        $this->lancamentoLoteRepository->deleteLancamentos($params);

        return true;
    }

    public function deleteLoteLancamento(
        int $nrLote
    ): bool {

        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRLOTE' => $nrLote,
            'P_NRORG'  => $nrOrg
        ];

        $this->deleteLancamentos($nrLote);
        $this->lancamentoLoteRepository->deleteLoteLancamento($params);

        return true;
    }
}
